
#include <inttypes.h>
#include <string.h>

/* setup.c
 * inspired by the example.c Linux bootloader code, distributed under BSD licence
 * found under http://www.simtec.co.uk/products/SWLINUX/files/booting_article.html
 */

/* list of possible tags */
#define ATAG_NONE       0x00000000
#define ATAG_CORE       0x54410001
#define ATAG_MEM        0x54410002
#define ATAG_VIDEOTEXT  0x54410003
#define ATAG_RAMDISK    0x54410004
#define ATAG_INITRD2    0x54420005
#define ATAG_SERIAL     0x54410006
#define ATAG_REVISION   0x54410007
#define ATAG_VIDEOLFB   0x54410008
#define ATAG_CMDLINE    0x54410009

/* structures for each atag */
struct AtagHeader {
	uint32_t size; /* length of tag in words including this header */
	uint32_t tag;  /* tag type */
};

struct AtagCore {
	uint32_t flags;
	uint32_t pagesize;
	uint32_t rootdev;
};

struct AtagMem {
	uint32_t     size;
	uint32_t     start;
};

struct AtagVideotext {
	uint8_t x;
	uint8_t y;
	uint16_t video_page;
	uint8_t video_mode;
	uint8_t video_cols;
	uint16_t video_ega_bx;
	uint8_t video_lines;
	uint8_t video_isvga;
	uint16_t video_points;
};

struct AtagRamdisk {
	uint32_t flags;
	uint32_t size;
	uint32_t start;
};

struct AtagInitrd2 {
	uint32_t start;
	uint32_t size;
};

struct AtagSerialnr {
	uint32_t low;
	uint32_t high;
};

struct AtagRevision {
	uint32_t rev;
};

struct AtagVideolfb {
	uint16_t lfb_width;
	uint16_t lfb_height;
	uint16_t lfb_depth;
	uint16_t lfb_linelength;
	uint32_t lfb_base;
	uint32_t lfb_size;
	uint8_t red_size;
	uint8_t red_pos;
	uint8_t green_size;
	uint8_t green_pos;
	uint8_t blue_size;
	uint8_t blue_pos;
	uint8_t rsvd_size;
	uint8_t rsvd_pos;
};

struct AtagCmdline {
	char cmdline[0];
};

struct Atag {
	struct AtagHeader hdr;
	union {
		struct AtagCore core;
		struct AtagMem mem;
		struct AtagVideotext videotext;
		struct AtagRamdisk ramdisk;
		struct AtagInitrd2 initrd2;
		struct AtagSerialnr serialnr;
		struct AtagRevision revision;
		struct AtagVideolfb videolfb;
		struct AtagCmdline cmdline;
	};
};


#define tagNext(t) ((struct Atag *)((uint32_t *)(t) + (t)->hdr.size))
#define tagSize(type)  ((sizeof(struct AtagHeader) + sizeof(struct type)) >> 2)

static void setupCoreTag(struct Atag ** tag, long pagesize) {
	(*tag)->hdr.tag = ATAG_CORE;
	(*tag)->hdr.size = tagSize(AtagCore);
	
	(*tag)->core.flags = 1;
	(*tag)->core.pagesize = pagesize;
	(*tag)->core.rootdev = 0;
	
	*tag = tagNext(*tag);
}

static void setupRamdiskTag(struct Atag ** tag, uint32_t size) {
	(*tag)->hdr.tag = ATAG_RAMDISK;
	(*tag)->hdr.size = tagSize(AtagRamdisk);
	
	(*tag)->ramdisk.flags = 0;
	(*tag)->ramdisk.size = size;
	(*tag)->ramdisk.start = 0;
	
	*tag = tagNext(*tag);
}

static void setupInitrd2Tag(struct Atag ** tag, uint32_t start, uint32_t size) {
	(*tag)->hdr.tag = ATAG_INITRD2;
	(*tag)->hdr.size = tagSize(AtagInitrd2);
	
	(*tag)->initrd2.start = start;
	(*tag)->initrd2.size = size;
	
	*tag = tagNext(*tag);
}

static void setupMemTag(struct Atag ** tag, uint32_t start, uint32_t size) {
	(*tag)->hdr.tag = ATAG_MEM;
	(*tag)->hdr.size = tagSize(AtagMem);
	
	(*tag)->mem.start = start;
	(*tag)->mem.size = size;
	
	*tag = tagNext(*tag);
}

static void setupCmdlineTag(struct Atag ** tag, const char * line) {
	int len = strlen(line);
	
	(*tag)->hdr.tag = ATAG_CMDLINE;
	(*tag)->hdr.size = (sizeof(struct AtagHeader) + len + 1 + 4) >> 2;
	
	strcpy((*tag)->cmdline.cmdline, line);
	
	*tag = tagNext(*tag);
}

static void setupEndTag(struct Atag ** tag) {
	(*tag)->hdr.tag = ATAG_NONE;
	(*tag)->hdr.size = 0;
}

void linuxSetupParameters(void * loc, uint32_t ramStart, uint32_t ramLen,
	uint32_t initrdLoad, uint32_t initrdLen) {
	struct Atag * param = (struct Atag*)loc;
	
	setupCoreTag(&param, 4096); // 4k pagesize
	setupMemTag(&param, ramStart, ramLen);
	//setupMemTag(&param, 0x5c000000, 256 * 1024);
	setupInitrd2Tag(&param, initrdLoad, initrdLen);
	//setupCmdlineTag(&param, "console=ttyS2,115200 root=/dev/ram0");
	setupEndTag(&param);
}

